/*
 *  DVDSubEdit and this code is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVDSubEdit and this code are distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
*/

#include "stdafx.h"
#include "FixVTS.h"
#include "FixVTSDlg.h"
#include "Cell.h"
#include "IFOUtils.h"
#include "Utils.h"
#include <mmsystem.h>
#include <math.h>

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

void OutputError(char* AMessage, int Langage)
{
	AfxMessageBox(AMessage);
#if _DEBUG
	OutputDebugString(Message);
	OutputDebugString("\n");
#endif
}


/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();
	CVTSClean*	m_Cleaner;

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL


// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	virtual BOOL OnInitDialog();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()


BOOL CAboutDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();
	SetDlgItemText(IDC_VERSION,m_Cleaner->m_Version);
	return TRUE; 
}


/////////////////////////////////////////////////////////////////////////////
// CFixVTSDlg dialog

CFixVTSDlg::CFixVTSDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CFixVTSDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CFixVTSDlg)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);

	pErrorFunction = OutputError;
	m_CommandLineMode = false;
	m_CMDInputFile = 0;
	m_CleanupBackup = 0;
	m_InPlaceProcess = 0;
	m_StartRightAway = 0;
	m_ProcessFullDVD = 0;
	m_RetVal = 0;
	m_DefaultLogPath = true;
}

void CFixVTSDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CFixVTSDlg)
	DDX_Control(pDX, IDC_BOGUS, m_BogusButton);
	DDX_Control(pDX, IDC_INPLACE, m_InPlaceButton);
	DDX_Control(pDX, IDC_PROGRESS, m_Progress);
	DDX_Control(pDX, IDC_UNREF, m_UnrefButton);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CFixVTSDlg, CDialog)
	//{{AFX_MSG_MAP(CFixVTSDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDC_RUN, OnRun)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_UNREF, OnUnref)
	ON_BN_CLICKED(IDC_BOGUS, OnBogus)
	ON_WM_CTLCOLOR()
	ON_BN_CLICKED(IDC_RESTORE, OnRestore)
	ON_BN_CLICKED(IDC_DELETE, OnDeleteBackup)
	ON_BN_CLICKED(IDC_PAUSE, OnPause)
	ON_BN_CLICKED(IDC_INPLACE, OnInplace)
	ON_WM_CLOSE()
	ON_BN_CLICKED(IDC_RUNFULL, OnRunfull)
	ON_BN_CLICKED(IDC_OPEN, OnOpen)
	//}}AFX_MSG_MAP
	ON_NOTIFY_EX( TTN_NEEDTEXT, 0, OnToolTipNotify )
	ON_WM_DROPFILES()
END_MESSAGE_MAP()

BOOL CFixVTSDlg::OnToolTipNotify( UINT id, NMHDR * pNMHDR, LRESULT * pResult )
{
	// Multiline tooltips: http://www.thecodeproject.com/miscctrl/tooltipex.asp
    TOOLTIPTEXT *pTTT = (TOOLTIPTEXT *)pNMHDR;
    UINT nID =pNMHDR->idFrom;
    if (pTTT->uFlags & TTF_IDISHWND)
    {
        // idFrom is actually the HWND of the tool
        nID = ::GetDlgCtrlID((HWND)nID);
		switch(nID)
		{
		case IDC_UNREF : pTTT->lpszText = "Select to remove unreferenced vob cells and blank packs.\
If unchecked, blank packs are replaced by suffing packs"; break; 
		case IDC_BOGUS : pTTT->lpszText = "Select to fix start/end sector pointers in PGC when they're clearly wrong: Highly recommended"; break; 
		case IDC_RUN : pTTT->lpszText = "Start cleaning the VOB file you dragged/dropped"; break; 
		case IDC_RUNFULL : pTTT->lpszText = "Process all the DVD VOBs for a full cleanup"; break; 
		case IDC_RESTORE : pTTT->lpszText = "Restore backups"; break; 
		case IDC_DELETE : pTTT->lpszText = "Delete backup files in FixVTSBackup"; break; 
		case IDC_INPLACE : pTTT->lpszText = "Process in-place, without backing up the VOB files.\
Usually much faster"; break; 
		}
		pTTT->hinst = AfxGetResourceHandle();
		return(TRUE);
    }
    return(FALSE);
}


/////////////////////////////////////////////////////////////////////////////
// CFixVTSDlg message handlers

BOOL CFixVTSDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_STRING, SC_MINIMIZE, "&Minimize");
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX+0x10, "Open LOG");
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here
	DragAcceptFiles();

	TheSettings.LoadSettings(m_hWnd,"MainWindowPos",1);
	if(!m_CommandLineMode)
	{
		TheSettings.LoadSettings(&m_RemoveUnrefed,"RemoveUnrefed",1);
		TheSettings.LoadSettings(&m_RemapVOBID,"RemapVOBIDs",1);
		TheSettings.LoadSettings(&m_InPlaceProcess,"InPlaceProcess",0);
		TheSettings.LoadSettings(&m_FixPGCLBAPtrs,"FixPGCLBAs",1);
	}
	m_UnrefButton.SetCheck(m_RemoveUnrefed);
	m_InPlaceButton.SetCheck(m_InPlaceProcess);
	TheSettings.LoadSettings(&m_PutBackupAbove,"BackupAbove",m_PutBackupAbove);
	m_BogusButton.SetCheck(m_FixPGCLBAPtrs);
	InitializeCriticalSection(&m_csLock);
	
	EnableToolTips(true);
	m_QuitNow = 0;
	m_ThreadHandle = 0;
	m_WarnInPlace = true;
	if(m_CommandLineMode && !m_StartRightAway)
		SetDlgItemText(IDC_MESSAGE,"To start, press Cleanup");
	else
		SetDlgItemText(IDC_MESSAGE,"To start, drag/drop a problematic VOB file");
	SetWindowText(m_Version);

	// Log path.
	if(m_DefaultLogPath)
	{
		if(getenv("TEMP") != 0)
			sprintf(m_LogPath,"%s\\FixVTS.log",getenv("TEMP"));
		else
			sprintf(m_LogPath,"C:\\temp\\FixVTS.log");
	}

	if(m_CommandLineMode && m_CMDInputFile)
	{
		m_RetVal = -OpenFile(m_CMDInputFile);
		if(m_RetVal != 0)
			AfxGetMainWnd()->SendMessage(WM_CLOSE);
		else
		{
			if(m_CleanupBackup)
				DeleteBackup();
			if(m_StartRightAway)
				OnRun();
		}
	}

	return TRUE;  // return TRUE  unless you set the focus to a control
}



int CFixVTSDlg::OpenFile(const char* FileName)
{
	int VTS,Index;
	if(sscanf(GetFname(FileName),"%*[^_]_%d_%d",&VTS,&Index) != 2 &&
		strcmpi(GetFname(FileName),"video_ts") != 0)
	{
		AfxMessageBox("FixVTS can only process title vobs (VTS_xx_[1-10].VOB)");
		return -1;
	}
	int ret = CVTSClean::OpenFile(FileName);
	if(ret == -1)
	{
		AfxMessageBox(m_CommandLineMode?"Couldn't open input file, exiting":"Please, drag/drop a VOB file");
		return -1;
	}
	GetDlgItem(IDC_RUN)->EnableWindow(true);
	GetDlgItem(IDC_RUNFULL)->EnableWindow(true);
	GetDlgItem(IDC_RESTORE)->EnableWindow(true);
	GetDlgItem(IDC_DELETE)->EnableWindow(true);
	SetWindowText(CatStr("Ready to process ",GetFname(m_FileName)));
	return 0;
}

void CFixVTSDlg::OnClose() 
{
	m_QuitNow = true;
	// If we're running, let's act as if the user had decided to abort processing.
	if(m_Running)
	{
		OnRun();
		// This would indicate that the user decided not to abort.
		if(!m_AbortNow) return;
	}
	WaitForSingleObject(m_ThreadHandle,1000);
	// Remember settings then quit.
	if(!m_CommandLineMode)
		TheSettings.SaveSettings();
	CDialog::OnClose();
}


UINT FixVTSProcessFunction( LPVOID pParam )
{
	CFixVTSDlg* Dialog = (CFixVTSDlg*) pParam;
	Dialog->m_Timer = Dialog->SetTimer(0,200,NULL);
	EnterCriticalSection(&Dialog->m_csLock);
	Dialog->DragAcceptFiles(FALSE);
	Dialog->SetDlgItemText(IDC_RUN,"Abort");
	Dialog->m_RemoveUnrefed = Dialog->m_UnrefButton.GetCheck();
	Dialog->m_FixPGCLBAPtrs = Dialog->m_BogusButton.GetCheck();
	Dialog->GetDlgItem(IDC_PAUSE)->ShowWindow(SW_SHOW);
	Dialog->GetDlgItem(IDC_OPEN)->EnableWindow(FALSE);
	if(Dialog->m_ProcessFullDVD)
		Dialog->SetWindowText("Processing full DVD");

	if(!Dialog->m_ProcessFullDVD)
		Dialog->Process();
	else
		Dialog->ProcessFullDVD();

	if(!Dialog->m_QuitNow)
	{
		Dialog->m_ProcessFullDVD = 0;
		Dialog->m_Pause = 0;
		Dialog->m_Running = 0;
		Dialog->KillTimer(Dialog->m_Timer);
		Dialog->DragAcceptFiles(TRUE);
		Dialog->SetDlgItemText(IDC_RUN,"Cleanup");
		Dialog->GetDlgItem(IDC_OPEN)->EnableWindow(TRUE);
		Dialog->EnableUI(true);
		Dialog->SetWindowText(CatStr("Done processing ",GetFname(Dialog->m_FileName)));
		Dialog->GetDlgItem(IDC_PAUSE)->ShowWindow(SW_HIDE);
		Dialog->SetDlgItemText(IDC_PAUSE,(Dialog->m_Pause)?"Continue":"Pause");
		Dialog->m_Progress.SetPos(0);
	}
	CTime CurTime = CTime::GetCurrentTime();
	CString STStr = CurTime.Format("End Processing at %H:%M:%S\n");
	FILE* Outlog = fopen(Dialog->m_LogPath,"a");
	if(Outlog)
	{
		fprintf(Dialog->m_Outlog,STStr);
		fprintf(Dialog->m_Outlog,"__________________________________________________________\n");
		fclose(Outlog);
	}
	if(!Dialog->m_InPlaceProcess && (Dialog->m_AbortNow || Dialog->m_QuitNow))
	{
		// If the user aborts , let's restore the backup!
		Dialog->RestoreBackup();
		Dialog->m_RetVal = 1;
	}
	if(Dialog->m_CommandLineMode)
	{
		if(Dialog->m_StartRightAway)
			AfxGetMainWnd()->SendMessage(WM_CLOSE);
	}
	LeaveCriticalSection(&Dialog->m_csLock);
	Dialog->m_ThreadHandle = 0;
	return 0;
}


void CFixVTSDlg::OnRunfull() 
{
	m_ProcessFullDVD = true;
	OnRun();
}

void CFixVTSDlg::OnRun() 
{

	if(m_Running == 0)
	{
		if(m_InPlaceProcess && !m_CommandLineMode && m_WarnInPlace && 
AfxMessageBox("Warning: you're about to process in-place, without backing up\n\
the VOB files. FixVTS will not damage the files, even if you abort, but you will not be\n\
able to recover their initial state. This is potentially much faster, however...\n\
Do you want to proceed?",MB_OKCANCEL) != IDOK)
			return;
		m_AbortNow = 0;
		EnableUI(false);
		m_Outlog = fopen(m_LogPath,m_DefaultLogPath ? "w" : "a");
		CTime CurTime = CTime::GetCurrentTime();
		CString STStr = CurTime.Format("Starting Processing at %H:%M:%S");
		MESSAGE1(STStr);
		CWinThread *Temp = AfxBeginThread( FixVTSProcessFunction, this);
		m_ThreadHandle = Temp->m_hThread;
		m_Running = 1;
		if(m_InPlaceProcess)
			m_WarnInPlace = 0;
	}
	else
	{
		m_Pause = 1;
		if(m_InPlaceProcess && (m_NumPacksRemoved > 0 || m_NumPacksRemapped > 0) && 
AfxMessageBox("Warning: you're about to abort in-place processing. \n\
Some packs have already been removed or renumbered, but the IFO files have not\n\
been updated yet. This would leave the DVD in a very messy state, you should click No!.\n\
Do you still want to abort processing?",MB_YESNO) != IDYES)
		{
			m_Pause = 0;
			return;
		}
		m_Pause = 0;
		m_AbortNow = 1;
	}
}

void CFixVTSDlg::EnableUI(bool Enable)
{
	GetDlgItem(IDC_UNREF)->EnableWindow(Enable);
	GetDlgItem(IDC_BOGUS)->EnableWindow(Enable);
	GetDlgItem(IDC_RESTORE)->EnableWindow(Enable);
	GetDlgItem(IDC_DELETE)->EnableWindow(Enable);
	GetDlgItem(IDC_INPLACE)->EnableWindow(Enable);
	GetDlgItem(IDC_RUNFULL)->ShowWindow(Enable?SW_SHOW:SW_HIDE);
}

void CFixVTSDlg::OnTimer(UINT nIDEvent) 
{
	// Update progress bar...
	DWORD ThisTime = timeGetTime();
	FLOAT RemainingSec = 0;
	static DWORD LastTime  = 0;
	static int LastInLBA = 0;
	static FLOAT BitRate = 20;
	static FLOAT BitRateSmooth = 20;
	static int CumulNumInputLBA = 0;
	if(m_Phase == Processing_VOB && ThisTime > LastTime + 2000 && !m_Pause)
	{
		// Compute the processing bitrate... CumulInPackIndex == m_CumulInPackIndex indicates
		// we're still processing the same vob.
		if(CumulNumInputLBA == m_CumulNumInputLBA)
		{
			float NewBitRate = 1000.0f*2*(m_CumulInPackIndex - LastInLBA) / (ThisTime - LastTime)/1024;
			if(fabs(BitRate - NewBitRate) > 4)
				BitRate = .0f*BitRate + 1.0f*NewBitRate;
			else
				BitRate = .98f*BitRate + .02f*NewBitRate;
			BitRateSmooth = .98f*BitRateSmooth + .02f*NewBitRate;
		}
		LastInLBA = m_CumulInPackIndex;
		LastTime = ThisTime;
		CumulNumInputLBA = m_CumulNumInputLBA;
	}
	RemainingSec = 2*(m_CumulNumInputLBA - m_CumulInPackIndex) / 1024 / BitRateSmooth;
	char Temp[512];
	switch(m_Phase)
	{
	case Processing_Idle : sprintf(Temp,"");break;
	case Processing_VOB : 		
		SetWindowText(CatStr("Processing ",GetFname(m_FileName)));
		if(!m_WaitForFile)
			sprintf(Temp,"%d%% Processed (%.2f MB/s) -- %d:%02d secs to go",
				100*m_CumulInPackIndex / (m_CumulNumInputLBA+1), BitRate,(int)RemainingSec/60,
				(int)RemainingSec%60);
		else
			sprintf(Temp,"%.2f MB/s",BitRate);

		m_Progress.SetPos(100*m_InPackIndex / (m_NumInputLBA+1));
		break;
	case Processing_IFO : sprintf(Temp,"Processing IFOs");break;
	case Processing_Done : sprintf(Temp,"Done...");break;
	case Processing_Aborted : sprintf(Temp,"Aborted");break;
	case Writing_IFO : sprintf(Temp,"Writing IFOs");break;
	}
	SetDlgItemText(IDC_INFO,Temp);
	SetDlgItemText(IDC_MESSAGE,Message);
	CDialog::OnTimer(nIDEvent);
}

//////////////////////////////////////////////////////////////////////////////////////////////
void CFixVTSDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.m_Cleaner = this;
		dlgAbout.DoModal();
	}
	else if ((nID & 0xFFF0) == IDM_ABOUTBOX+0x10)
	{
		ShellExecute(NULL, "open", m_LogPath, NULL, NULL, SW_SHOWNORMAL);
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CFixVTSDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CFixVTSDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CFixVTSDlg::OnDropFiles(HDROP hDropInfo) 
{
	SetForegroundWindow();      // activate us first !
	POINT ThePoint;
	DragQueryPoint(hDropInfo,&ThePoint);

	UINT nFiles = ::DragQueryFile(hDropInfo, (UINT)-1, NULL, 0);
	CWinApp* pApp = AfxGetApp();
	ASSERT(pApp != NULL);
	for (UINT iFile = 0; iFile < nFiles; iFile++)
	{
		TCHAR szFileName[_MAX_PATH];
		::DragQueryFile(hDropInfo, iFile, szFileName, _MAX_PATH);
		OpenFile(szFileName);
	}
	::DragFinish(hDropInfo);
}


void CFixVTSDlg::OnUnref() 
{
	m_RemoveUnrefed = m_UnrefButton.GetCheck();
}

void CFixVTSDlg::OnBogus() 
{
	m_FixPGCLBAPtrs = m_BogusButton.GetCheck();
}

void CFixVTSDlg::OnInplace() 
{
	m_InPlaceProcess = m_InPlaceButton.GetCheck();
}


HBRUSH CFixVTSDlg::OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor) 
{
	HBRUSH hbr = CDialog::OnCtlColor(pDC, pWnd, nCtlColor);
	int id = pWnd->GetDlgCtrlID();
	//COLORREF bkcolor = pDC->GetBkColor();
	if (id == IDC_INFO || id == IDC_MESSAGE)
	{
		// Set the text color to black
		pDC->SetTextColor(RGB(0, 0, 0)); 
	}	
	// TODO: Return a different brush if the default is not desired
	return hbr;
}

void CFixVTSDlg::OnRestore() 
{
	if(m_CommandLineMode)
	{
		RestoreBackup();
		return;
	}
	MessageBeep(MB_ICONEXCLAMATION);
	if(AfxMessageBox("This will restore your backup and undo the last procesing! Proceed?",MB_OKCANCEL) == IDOK)
		RestoreBackup();
}

void CFixVTSDlg::OnDeleteBackup() 
{
	if(m_CommandLineMode)
	{
		DeleteBackup();
		return;
	}
	MessageBeep(MB_ICONEXCLAMATION);
	if(AfxMessageBox("This will delete all backup files! Proceed?",MB_OKCANCEL) == IDOK)
		DeleteBackup();
}

void CFixVTSDlg::OnPause() 
{
	m_Pause = !m_Pause;
	SetDlgItemText(IDC_PAUSE,(m_Pause)?"Continue":"Pause");
}


void CFixVTSDlg::OnOpen() 
{
	// Pop an open dialog and get going!
	static CFileDialog FileDialog(TRUE, CString(".VOB"), NULL, OFN_HIDEREADONLY | OFN_OVERWRITEPROMPT, 
		CString("VOB Files (*.VOB)|*.VOB||"));
	FileDialog.m_ofn.lpstrTitle = "Load VOB file";
//	FileDialog.m_ofn.lpstrInitialDir = m_DefIFOPath;
	if (FileDialog.DoModal() == IDOK)
	{
		OpenFile(FileDialog.GetPathName());
	}
}
