#ifndef IFOUTILS_H
#define IFOUTILS_H

#include <stdio.h>
#include "includes.h"


// Main IFO structure.
typedef struct
{
	uchar	m_DVDVIDEO[12];
	uint	m_SetLastSector;
	uchar	m_Filler[12];
	uint	m_IFOLastSector;
	uchar	m_Head[0xc4-0x20];
	uint	m_PTT_SRPT;
	uint	m_PVTS_PTT_SRPT;
	uint	m_PVTS_PGCI;
	uint	m_PVTSM_PGCI_UT;
	uint	m_PVTS_TMAPTI;
	uint	m_PVTSM_C_ADT;
	uint	m_PVTSM_VOBU_ADMAP;
	uint	m_PVTS_C_ADT;
	uint	m_PVTS_VOBU_ADMAP;
	uchar	m_Tail[0x3d8-0xe4];
} IFO_main;

// items in the VTSM_C_ADT tables.
typedef struct
{
	ushort	m_VOBID;
	uchar	m_CELLID;
	uchar	m_Res;
	uint	m_StartLBA;
	uint	m_EndLBA;
} VTSM_C_ADTItem;

typedef struct
{
	ushort	m_NumVOBID;
	ushort	m_Reserved;
	uint	m_LastByte;
	VTSM_C_ADTItem	m_Items[1024*16];
} VTSM_C_ADTTable;

typedef struct
{
	uchar	m_Flags;
	uchar	m_Res;
	uchar	m_Still;
	uchar	m_Command;
	uint	m_PBTime;
	uint	m_FirstVOBUStartLBA;
	uint	m_FirstILVUEndLBA;
	uint	m_LastVOBUStartLBA;
	uint	m_LastVOBUEndLBA;
} CELL_PLAYBACK_INFO;

typedef struct
{
	ushort	m_VOBID;
	uchar	m_Res;
	uchar	m_CELLID;
} CELL_POS_INFO;

typedef struct
{
	ushort	m_Space;
	uchar	m_NumProg;
	uchar	m_NumCells;
	uint	m_PlaybackTime;
	uint	m_PUOs;
	ushort	m_AudioStreams[8];
	uint	m_SubpicStreams[32];
	ushort	m_NextPGC;
	ushort	m_PrevPGC;
	ushort	m_UpPGC;
	uchar	m_PGCStill;
	uchar	m_PlaybackMode;
	uint	m_Palette[16];
	ushort	m_OffsetToCommands;
	ushort	m_OffsetToPrograms;
	ushort	m_OffsetToCellPB;
	ushort	m_OffsetToCellPos;
} PGC_HEADER;

// Enum for types of PGC.
typedef enum
{
	PGCTypeFirstPlay = 0,
	PGCTypeNone,
	PGCTypeTitleMenu,
	PGCTypeRoot,
	PGCTypeSubpicture,
	PGCTypeAudio,
	PGCTypeAngle,
	PGCTypeChapter,
	PGCTypeTitle,
} PGCType;

// Audio attributes for the domain.
typedef struct
{
	uchar	m_CodingMode;
	uchar	m_Quantization;
	uchar	m_Language[2];
	uchar	m_Reserved;
	uchar	m_CodeExtension;
	uchar	m_Reserved2;
	uchar	m_KaraokeSurround;

} AUDIO_ATTRIBUTES;

// TT_SRPT is a collection of these title information.
typedef struct
{
	uchar	m_Type;
	uchar	m_NumAngles;
	ushort	m_NumChapters;
	ushort	m_ParentalMask;
	uchar	m_VTSNumber;
	uchar	m_TitleNumberInVTS;
	uint	m_VTSStartLBA;

} TT_SRPT_ENTRY;

typedef enum
{
	IFO_MAIN,			// Main IFO table.	
	VTSM_PGCI_UT,		// Table of PGC infos (menus).
	VTSM_C_ADT,			// Cell address table (menus).
	VTSM_VOBU_ADMAP,	// VOBU address table (menus).
	TT_SRPT,			// For manager only!
	FP_PGC,				// For manager only!
	VTS_C_ADT,			// Cell address table (titles).
	VTS_PGCI,			// Table of PGC infos (titles).
	VTS_VOBU_ADMAP,		// VOBU address table (titles).
	VTS_TMAPTI,
	VMGM_PGCI_UT,
	VTS_PTT_SRPT,		// Titles and chapters (NOT MANAGER).
	VMG_TXTDT_MG,
	VMG_PTL_MAIT,
	VMG_VTS_ATRT
} IFOTable;

// This structure can be used to define a mapping to renumber VCID.
typedef struct
{
	ushort	InVOBID;
	uchar	InCELLID;
	ushort	OutVOBID;
	uchar	OutCELLID;
} VCIDMapItem;

class CCell;

class CIFOLoader
{
public: 

	// Creates the class, according to an IFO file passed as an argument.
	CIFOLoader();
	CIFOLoader(int VTS);
	CIFOLoader(const char* IFOFileName);
	~CIFOLoader();

	// Loads the given IFO.
	int				LoadIFO(const char* IFOFileName);
	int				LoadIFO(int VTS);
	void			CloseIFO();

	// If you need direct access.
	uchar*			GetBuffer() {return m_Buffer;}

	// Write IFO back on to original if no filename. Returns 0 if no problem.
	int				WriteIFO(char* WritePath = NULL);

	// This finds this table and returns a pointer to it in m_Buffer.
	uchar*			FindTable(IFOTable Table);
	// This find the end address of the table. IncEndAddress() adds IncBytes to it. In bytes relative
	// to the start of the table.
	uint			GetEndAddress(IFOTable Table);
	void			IncEndAddress(IFOTable Table, int IncBytes);
	// This returns the absolute end address, relative to the start of the IFO file. In bytes.
	uint			GetAbsoluteEndAddress(IFOTable Table);

	// This checks whether the table size can change without moving other tables up or down.
	bool			TableSizeCanChange(IFOTable Table, int NumBytes);

	// Use this to change the table size. It will move other tables if needed. It does not increment
	// the end address for this table.
	void			ChangeTableSize(IFOTable Table, int NumBytes);
	
	// This removes the table entirely, moving what needs to be moved...
	void			RemoveTable(IFOTable Table);

	// Use this to move data up or down in table. It will resize the table accordingly and will modify
	// the end address for this table.
	void			MoveDataInTable(IFOTable Table, uchar* Ptr, int NumBytes);

	// These functions are for accessing the VTSM_PGCI_UT. *** All indexes start at 1 ***
	ushort			GetNumLU();
	ushort			GetNumPGC(ushort LU);
	uchar			GetNumCells(ushort LU, ushort PGC);
	uchar			GetNumPrograms(ushort LU, ushort PGC);
	void			SetNumPrograms(ushort LU, ushort PGC, uchar NumProg);
	uchar*			GetProgramTable(ushort LU, ushort PGC);

	// Use this to retrieve the number of VTS titles defined in this IFO.
	ushort			GetNumberVTSTitles();
	// Use this to retrieve the number of chapters for this title. title starts at 0 (first title).
	uint			GetNumberChapters(ushort title);
	// This removes the vts title title (starts at 1).
	void			RemoveVTSTitle(int title);
	// This removes the DVD title title (starts at 1). 
	void			RemoveDVDTitle(int title);

	// This removes the PGC from VTS(M)_PGCI_UT. Use LU=0 for title domain. ONLY LU==0 IS SUPPORTED.
	void			RemovePGC(ushort LU, ushort PGC);

	// This looks through all the LUS, and all the PGCS, to find all occurences of the cell with the
	// specified VOBID and CELLID. Returns the number of cells found.
	// LU, PGC and CELLOut must be arrays that are large enough to receive the results.
	int			FindThisCell(ushort VOBID, uchar CELLID, ushort *LU, ushort *PGC, ushort *CELLOut);
	// This simply tells you if the given VCID is referenced in any of the PGCs for the desired domain.
	bool		IsReferenced(ushort VOBID, uchar CELLID, bool Menu = true);
	// This looks through all the cells declared in C_ADT table, and returns true if any is unrefed.
	bool		AreAnyCellUnreferenced(bool menu = true);
	
	// This copies the CLUT from the PGC info into the array. Use LU=0 to look in the title domain.
	// CLUT will contain 3*16 uchars, representing Y, Cr, Cb for each element in the CLUT.
	void		GetCLUT(uchar *CLUT, ushort LU, ushort PGC, bool RGB = false);
	void		SetCLUT(uchar *CLUT, ushort LU, ushort PGC, bool RGB = false);

	// This returns the number of subpicture streams defined in the domain, and puts in
	// Languages a pair of characters for each language.
	int			GetSubpicLangages(char* Languages, bool Menu = true);
	// For the given LU (can be 0 for title domain) and PGC, and the given subpic StreamID,
	// this returns the 2 language char in Languages. Returns true if the stream was found.
	// Upon return, the bits in DispModes are set to indicate which display modes the streamID is
	// used for: 0 for 4:3, 1 for widescreen, 2 for LB and bit 3 for PS.
	bool		GetPgcSubpicLanguage(char* Languages, ushort LU, ushort PGC, uchar StreamID, uchar* DispModes = 0);

	// This tests whether the subpic stream is available. Index is between 0 and 31, corresponds to the domain attributes.
	bool		IsSubpicStreamAvailable(ushort LU, ushort PGC, uchar Index);

	// This goes through all the PGCs, and collects all the subpic streams used, then updates the IFO table accordingly.
	void		FixSubpicAttributes(bool menu = true);

	// This kills the playback of the pgc. Returns -1 if it can't 0 if all OK.
	int			BypassPGCPlayback(ushort LU, ushort PGC);

	// *** All indexes start at 1 ***
	CELL_PLAYBACK_INFO*		GetCellPlaybackInfo(ushort LU, ushort PGC, ushort Cell);
	CELL_POS_INFO*			GetCellPosInfo(ushort LU, ushort PGC, ushort Cell);
	PGC_HEADER*				GetPGCHeader(ushort LU, ushort PGC);
	// This returns the type of PGC (what kind of menu, or title number) if any...
	PGCType					GetPGCType(ushort LU, ushort PGC, uchar* TitleNumber = 0);
	// This allows you to get the next, previous and up PGC.
	void					GetNPUPGC(ushort LU, ushort PGC, ushort &NextPGC, ushort &PrevPGC, ushort &UpPGC);

	// Get the audio attributes for the domain. There are 8 of them. ** ID between 0 and 7 **
	AUDIO_ATTRIBUTES*		GetAudioAttributes(uint ID, bool menu = true);
	// Careful: this is a short.
	uchar*					GetNumAudioStreams(bool menu = true);
	// This goes through all the PGCs, and collects all the audio streams used, then updates the IFO table accordingly.
	void					FixAudioAttributes(bool menu = true);
	uchar*					GetVideoAttribute();
	// Use this to get info on the audio attributes available to 1 PGC. Any of the pointers can be null if you don't want
	// the info. Index is the audio index (0 to 7). Returns 0 if all OK, or -1 if the stream isn't available.
	int						GetPGCAudioInfo(ushort LU, ushort PGC, uchar Index, char* Lang, char* Encoding, 
								uchar* NumChans, uchar* StreamID);

	// Returns the number of commands. Use Type = 0 for pre, 1 for post, 2 for cell and 3 for total.
	ushort			GetNumCommands(ushort LU, ushort PGC, int Type);
	// Returns pointers to number of pre, post and cell commands, and also a pointer to them.
	uchar*			GetCommands(ushort LU, ushort PGC, int Type, uchar** Commands = 0);

	// These are for adding commands at the end of the commands. 
	// Returns -1 if failed (too many already) or the incremented number of commands.
	int				AddCommand(ushort LU, ushort PGC, uchar* Command, int Type);
	// This version inserts the command at the start of the commands;
	int				InsertCommand(ushort LU, ushort PGC, uchar* Command, int Type);
	// This removes CommandIndex from the commands. CommandIndex starts at 1.
	int				RemoveCommand(ushort LU, ushort PGC, int Type, int CommandIndex);

	// This will update the info relative to the Cell in the IFO, in all the PGCs that reference
	// that cell and in VTS(M)_C_ADT. If AddToVTSM_C_ADT is true and the cell isn't referenced in the 
	// VTSM_C_ADT table, the cell entry will be added to the table. Call this after each cell in the VTS.
	void			UpdateCellInfo(CCell* TheCell, bool AddToVTSM_C_ADT = false, bool Menu = true);

	// This version only updates the VTS(M)_C_ADT table. IfAddToVTSM_C_ADT is true and the cell isn't 
	// referenced in the VTSM_C_ADT table, the cell entry will be added to the table.
	void			UpdateVTS_C_ADT(CCell* TheCell, bool AddToVTSM_C_ADT = false, bool Menu = true);

	// Use this to remove this cell from the VTS(M)_C_ADT table (for example, if it's unreferenced).
	void			RemoveFromVTS_C_ADT(CCell* TheCell, bool Menu = true);

	// This will go through all the cells in VTS(M)_C_ADT and check whether they're reference. It will
	// remove any that isn't.
	void			CleanupVTS_C_ADT(bool Menu = true);

	// This reconstructs VTS_C_ADT from scratch, from a list of VTSM_C_ADTItem.
	void			RebuildVTS_C_ADT(VTSM_C_ADTItem *Items, int NumCells, bool Menu = true);
	// From a list of VTSM_C_ADTItem, this also corrects the PGC playback tables.
	void			UpdateCellInfo(VTSM_C_ADTItem *Items, int NumCells, bool Menu = true);

	// This version does not use a CCell, rather two arrays of ints indicating the before/after navpack 
	// LBAs. The PGC playback info is updated, as well as the VOBUADMAP table. This should be called once
	// all the navpacks in the VTS have been processed. NOTE: By convention, the last navpack in the
	// arrays NEEDS TO POINT to the first LBA past the end of the VOB. So the real number of navpacks
	// in the vob is actually NumNavPacks - 1;
	void			UpdateNavPackInfo(int *InNavPacks, int *OutNavPacks, int NumNavPacks, bool Menu = true);

	// This will add the times in all the Cells in this PGC, and update the PGC time. Use LU = 0 if in the
	// title domain. If FixZeroTimes is true, any cell that has an elapsed time of 0 is changed to a valid value.
	// If AdjustFormat is true, the current value of the global VideoFormat variable is used in the time format.
	void			UpdatePGCPlayBackTime(ushort LU, ushort PGC, bool FixZeroTimes = false, bool AdjustFormat = false);

	// This updates the table of VOBUS, from the data gathered during the calls to UpdateCellInfo. Call
	// this after calling UpdateCellInfo() for every cell in the VTS.
	void			UpdateVOBUADMAP(bool Menu = true);

	// This version simply uses an array...
	void			UpdateVOBUADMAP(uint* StartLBA, int Num, bool Menu = true);

	// This updates the time-map table based on an input-output correspondence table. I don't like it
	// one bit.
	void			UpdateTMAPTI(int *InNavPacks, int *OutNavPacks, int NumNavPacks);

	// This can be used to renumber the VCIDs in an entire domain. It goes through all LUs (for menu)
	// all PGCs and all cells, and changes the VCID according to the Map. It does not change VTS(M)_C_ADT.
	void			RenumberVCID(VCIDMapItem* Map, int MapLength, bool Menu = true);

	// This updates the various sector pointers to the end of the IFO, beginning of the title vob etc.
	// This should be called on each VTS as the last step, with the video manager one as the last one.
	// If CorrectNumVTS is true and video_ts.ifo is opened, the actual number of VTS present on the DVD is reflected
	// in the tables (VMG_PTL_MAIT, VMG_VTS_ATRT and TT_SRPT).
	void			UpdateVTSSectors(bool CorrectNumVTS = false);

	// This can only be used when opening video_ts.ifo. It updates the number of VTS in various tables.
	void			UpdateNumVTS(int NumVTS);

	// This is true if the IFO loaded OK...
	bool			m_LoadedOK;

	// This returns the 1 byte prohibited region mask.
	uchar			GetRegionCodeMask();
	
	// This returns a pointer to VTSM_LU for the given LU.
	uchar*			GoToLU(ushort LU);

	// This goes through all the tables and fixes their "end byte" value.
	void			FixEndOfTables();
	
private:

	char			m_IFOFileName[1024];
	int				m_VTSNum;		// Which VTS is this? 0 for manager.

	FILE*			m_IFOfd;		// File descriptor for IFO file.
	long			m_IFOSizekB;
	long			m_MaxIFOSize;
	bool			m_IsManager;	// Is this video_ts.ifo?
	uchar*			m_Buffer;

	uchar*			GoToPCG(ushort LU, ushort PGC);
	uchar*			GoToCommands(ushort LU, ushort PGC);

	uint*			m_VOBUStartLBA;	// List of start LBAs for all the VOBUs.
	int				m_NumVOBUs;		// How many VOBUs in the VOB.

	void*			m_MinAddress;	// Points to start of IFO.
	void*			m_MaxAddress;	// Points to end of IFO.
	bool			m_AddressError;	// Normally false. Raised if address error.

	void			PutInt(uchar* Buffer, int Offset, uint Value);
	void			PutShort(uchar* Buffer, int Offset, ushort Value);
	uint			GetInt(uchar* Buffer, int Offset);
	ushort			GetShort(uchar* Buffer, int Offset);
	void			AddShort(uchar* Buffer, int Offset, short IncValue);
	void			AddInt(uchar* Buffer, int Offset, int IncValue);

	int				FixOffset(uchar* Ptr, ushort Skip=4);

};

#endif